#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include "id3.h"
#include "util.h"

static int verbose = 0;

extern int errno;

static char* output_text_according_to_encoding(cbyte encoding, cbyte* text, int length, FILE* out);
static struct id3v2_header* add_header(MP3Info* i);


void read_id3_header_verbosity(int v) 
{
    verbose = v;
}

MP3Info *read_id3_headers(MP3Info *mp3info,
                          const int read_tag_contents)
{
    size_t header_size;         /* total size of ID3 header (ie including header itself) */
    size_t taglen;
    uint8_t id3_flags;
    cbyte *p;
    cbyte *end_header;
    int i;
    
    p = mp3info->base;
    
    if (strncmp((const char*)p, "ID3", 3) != 0) {
        return mp3info;         /* JUMP OUT */
    }

    mp3info->major_version = p[3];
    mp3info->minor_version = p[4];

    id3_flags = p[5];
    if (id3_flags != 0) {
        fprintf(stderr, "%s: no flags currently supported, but set flags found (0x%x; implementation needs to be extended)\n", program_name(), id3_flags);
        exit(1);
    }
    mp3info->flags = id3_flags;

    header_size = 0;
    for (i=6; i<=9; i++) {
        header_size = header_size*128 + p[i]; /* only 7 bits significant */
    }

    header_size += 10; /* fix header_size: size in file doesn't include header */
    // mp3info->audio and audio_end were set when the file was first
    // mapped, but ->audio now needs to be updated
    mp3info->audio = &p[header_size];

    if (! read_tag_contents) {
        // We're not being asked to parse the contents of the ID3
        // header, but simply to find the endp pointer.  So jump out now.
        return mp3info;         /* JUMP OUT */
    }

    taglen = 0;                 /* dummy initial value */
    end_header = p+header_size;
    for (p += 10; p<end_header; p += taglen) {
        // p points to the beginning of the frame:
        // Frame ID       $xx xx xx xx (four characters) 
        // Size           $xx xx xx xx
        // Flags          $xx xx
        struct id3v2_header *header_current;

        if (p[0] == 0) {
            // this is header padding
            mp3info->padding = end_header - p;
            break;
        }

        header_current = add_header(mp3info);

        assert(mp3info->metadata != NULL);
        assert(header_current != NULL);

        taglen = 0;
        for (i=4; i<8; i++)            
            taglen = taglen*256 + p[i];

        header_current->id = (char*)p;
        header_current->size = taglen;
        header_current->flags = p[8]*0xff + p[9];
        header_current->deleted = 0;
        header_current->value = &p[10];

        taglen += 10;           /* account for 10 bytes of tag frame header */
    }

    if (mp3info->metadata == NULL) {
        fprintf(stderr, "In file <%s>, there is an ID3v2 header, but no metadata frames (invalid ID3 file; adding a dummy header)\n", mp3info->filename);
        struct id3v2_header *dummy;
        char *value = "\0Invalid ID3v2 file: no headers";
        dummy = add_header(mp3info);
        dummy->id = "TIT1";
        dummy->size = strlen(value+1) + 1;
        dummy->deleted = 0;
        dummy->value = (unsigned char*)value;
    }

    assert(mp3info->base != NULL);
    assert(mp3info->metadata != NULL);
    
    return mp3info;
}

void print_id3_frame(struct id3v2_header* h, FILE *output) 
{
    const char* id = h->id;
    cbyte* tagvalue = h->value;
    size_t taglen = h->size;

    if (h->deleted) {
        return;                 /* JUMP OUT */
    }
    
    fprintf(output, "%.4s=", id);

    if (strncmp(id, "APIC", 4) == 0) {
        // <Header for 'Attached picture', ID: "APIC">
        // Text encoding   $xx
        // MIME type       <text string> $00
        // Picture type    $xx
        // Description     <text string according to encoding> $00 (00)
        // Picture data    <binary data>
        cbyte* type;
        cbyte* description;
        for (description = &tagvalue[1]; *description != '\0'; description++)
            ;           /* go to the end of the MIME type string */
        type = ++description;
        description++;
        fprintf(output,
                "%s (type %d, %s, 0x%lx bytes)\n",
                (*description == '\0' ? (cbyte*)"<no description>" : description),
                *type,
                &tagvalue[1],
                taglen);
    } else if (strncmp(id, "COMM", 4) == 0) {
        // <Header for 'Comment', ID: "COMM">
        // Text encoding           $xx
        // Language                $xx xx xx
        // Short content descrip.  <text string according to encoding> $00 (00)
        // The actual text         <full text string according to encoding>
        cbyte* description = &tagvalue[4];
        // ignore the encoding and language for now
        printf("%s|", description);
        while (*description != '\0') {
            description++;
        }
        description++;
        fwrite(description, 1, taglen-(description-tagvalue), output);
    } else if (strncmp(id, "USLT", 4) == 0) {
        // <Header for 'Unsynchronised lyrics/text transcription', ID: "USLT"> 
        // Text encoding       $xx
        // Language            $xx xx xx
        // Content descriptor  <text string according to encoding> $00 (00)
        // Lyrics/text         <full text string according to encoding>
        cbyte encoding = tagvalue[0];
        output_text_according_to_encoding(encoding, &tagvalue[4], 0, output);
    } else if (id[0] == 'T') {
        // FIXME: there are other fields that need this call
        output_text_according_to_encoding(tagvalue[0], &tagvalue[1], taglen-1, output);
        //fputs(get_text_according_to_encoding(tagvalue, taglen), stdout);
    } else {
        int i;
        for (i=0; i<taglen; i++) {
            fprintf(output, "%c", tagvalue[i]);
        }
    }

    fputc('\n', output);

    return;
}

static char* output_text_according_to_encoding(cbyte encoding,
                                               cbyte* text,
                                               int length,
                                               FILE* out)
{
    if (encoding == 0) {
        // ASCII
        if (length == 0) {
            fputs((const char*)text, out);
        } else {
            fwrite(text, 1, length, out);
        }
    } else if (encoding == 1) {
        // UCS-2
        // XXX THE FOLLOWING DOESN'T WORK.  I can't work out how to
        // convert from UCS-2 to ascii, the big/little-endian
        // characters in, for example, USLT.  These appear as 
        //   000000c0  20 74 6f 20 74 68 65 20  66 69 6e 61 6c 55 53 4c  | to the finalUSL|
        //   000000d0  54 00 00 00 f8 00 00 01  65 6e 67 44 00 65 00 73  |T.......engD.e.s|
        //   000000e0  00 63 00 72 00 69 00 70  00 74 00 69 00 6f 00 6e  |.c.r.i.p.t.i.o.n|
        //   000000f0  00 00 00 54 00 69 00 6d  00 20 00 56 00 69 00 63  |...T.i.m. .V.i.c|
        //   00000100  00 6b 00 65 00 72 00 79  00 20 00 61 00 6e 00 64  |.k.e.r.y. .a.n.d|
        // ...so byte-null, byte-null, ....  I need to work out how to
        // tell mbstowcs to recognise these for what they are.
        //
        // I _think_ the way to do this is to take pairs of bytes,
        // multiply them into a char16_t integer, call c16rtomb to
        // convert this to a (few) bytes representing one mb
        // character, then call mbtowc to convert those bytes to a wchar_t.
        //
        // See test/other/test02.mp3
        //
        // ...but OS X 10.6.8 doesn't appear to have c16rtomb!

        // just hope for the best -- the result here isn't that
        // complete, but it doesn't actually break
        // (the following is safe, because the UCS-2 content has lots of nulls in it)
        fputs((const char*)text, out);
#if 0
#define NWCHAR 64
        wchar_t wchars[NWCHAR];
        int nwchars;
        char16_t c16;
        char mb[4];             /* I think that 4 bytes is enough for any multibyte character */
        cbyte* textp = text;

        // How do I tell the difference between little- and big-endian, here?
        // These strings can start with the BOMs: 0xff 0xfe, or 0xfe 0xff.
        nwchars = 0;
        while (nwchars < NWCHAR) {
            c16 = *textp;
            textp++;
            c16 += 0xff * *textp;
            textp++;
            c16rtomb(mb, c16, NULL);
            mbtowc(&wchars[nwchars], mb, 4);
            if (wchars[nwchars] == L'\0') {
                break;
            }
            if (length != 0 && (textp - text) > length) {
                break;
            }
        }
        if (nwchars >= NWCHAR) {
            fprintf(stderr, "output_text_according_to_encoding: buffer too small\n");
            wchars[NWCHAR-1] = L'\0';
        }
        
        fputws(wchars, out);
#endif

    } else {
        fprintf(stderr, "%s: Ooops: found illegal encoding %d\n",
                program_name(), encoding);
        exit(1);
    }
    return 0;
}

#if 0
static char* get_text_according_to_encoding(cbyte* p, int length)
{
    static char* buf = NULL;
    static int buflen = 0;

    int encoding = p[0];
    if (encoding != 0) {
        fprintf(stderr, "Ooops: found encoding %d: not ASCII (need to fix mp3x)\n", encoding);
    }
    // but carry on regardless, and see what happens...

    if (buf == NULL || length > buflen-1) {
        if (buf != NULL) {
            // not first time
            free(buf);
        }
        buflen = 3*length / 2 + 1;
        if ((buf = (char*)malloc(buflen)) == NULL) {
            fprintf(stderr, "Can't allocate buffer with length %d!\n", buflen);
            exit(1);
        }
    }
    // FIXME: right now, we ignore the encoding, and just assume everything's ASCII!
    strncpy(buf, (const char*)&p[1], length-1);
    buf[length-1] = '\0';
    return buf;
}
#endif

/*
 * Add a text field (ie 'Txxx').  The payload is an ASCII string, and
 * it's always encoded as ISO-8859-1.
 * If 'payload' is NULL, then the corresponding tag is removed rather
 * than changed.
 */
int add_id3_header_text(MP3Info* headers,
                        const char* tag,
                        const char* payload)
{
    struct id3v2_header* h;
    /* struct id3v2_header* h0 = NULL; */
    char* buf;
    size_t payload_length;
    int nh, i;
    
    if (tag[0] != 'T') {
        fprintf(stderr, "Can't add non-text headers: %.4s\n", tag);
        return 1;
    }
    
    h = headers->metadata;
    nh = headers->nmetadata;
    for (i=0; i<nh; i++) {
        if (h[i].deleted) {
            continue;
        }
        if (strncmp(h[i].id, tag, 4) == 0) {
            // match
            fprintf(stderr, "Deleting tag %s\n", tag);
            h[i].deleted = 1;
        }
    }

    // if payload==NULL, then we're to delete this tag, and there's
    // nothing more to do.

    if (payload != NULL) {
        payload_length = strlen(payload);
        if ((buf = (char*)malloc(4 + payload_length + 1)) == NULL) {
            /* +1 includes encoding byte */
            fprintf(stderr, "add_id3_header: Can't allocate %zd of memory for buffer\n",
                    4 + payload_length + 1);
            exit(1);
        }
        memcpy(buf, tag, 4);
        buf[4] = 0;                 /* encoding 0 => ISO-8859-1 */
        memcpy(&buf[5], payload, payload_length);

        h = add_header(headers);
    
        h->id = buf;
        h->size = payload_length + 1; /* +1 for encoding byte */
        h->flags = 0;                 /* always? */
        h->deleted = 0;
        h->value = (cbyte*)&buf[4];
    }

    return 0;
}

const MP3Info* write_header_and_file(const MP3Info* id3,
                                     const char* required_new_filename,
                                     const char** actual_new_filename)
{
    size_t body_size;
    size_t header_size;
    size_t id3_size;
    size_t new_padding_size = 0; /* only used if we write a new file */

    char* B;
    char* Bp;
    struct id3v2_header *h;
    int hn, i;
    const char* new_filename;

    const MP3Info *return_mp3;

    
    header_size = id3->audio - id3->base;
    body_size = id3->audio_end - id3->audio;
    id3_size = 0;               // we'll adjust this in a moment

    h = id3->metadata;
    hn = id3->nmetadata;
    for (i=0; i<hn; i++) {
        if (! h[i].deleted) {
            id3_size += (10 + h[i].size);
        }
    }
    // id3_size doesn't include the ID3 tag header itself, nor the padding

    if (id3_size == 0) {
        // Ooops: we're about to write an invalid ID3 header, with no
        // metadata.  We _could_ recover from this by writing dummy
        // headers, but this is a user error, so we should complain.
        fprintf(stderr, "%s: You appear to have deleted all of the tags in the ID3 header.\nThat's invalid, so I'm not going to write anything.\n", program_name());
        exit(1);
    }

    if ((B = (char*)malloc(id3_size+10)) == NULL) {
        fprintf(stderr, "Can't allocate %zd bytes for new header\n", id3_size+10);
        exit(1);
    }
    
    if (required_new_filename != NULL) {
        new_filename = required_new_filename;
        new_padding_size = 1024; /* arbitrary but generous */
    } else if (id3_size <= header_size) {
        // there's room to let us simply copy the new header into place
        new_filename = NULL;
        new_padding_size = 0;   /* we don't use this value below */
    } else {
        new_filename = tmpnam(NULL);
        new_padding_size = 1024; /* arbitrary but generous */
    }

    strncpy(B, "ID3", 3);
    B[3] = 3;
    B[4] = 0;
    B[5] = id3->flags;
    {
        int j;
        size_t size = id3_size + new_padding_size;
        for (j=9; j>=6; j--) {
            B[j] = (cbyte)(size & 0x7f);
            size >>= 7;
        }
    }
    
    Bp = &B[10];
    for (i=0; i<hn; i++) {
        if (h[i].deleted) {
            continue;
        }
        strncpy(Bp, h[i].id, 4);
        {
            /* According to
               http://www.hydrogenaudio.org/forums/index.php?showtopic=67145
               these are _not_ 'syncsafe' integers in ID3v2.3 */
            int j;
            size_t size = h[i].size;
            for (j=7; j>=4; j--) {
                Bp[j] = (cbyte)(size & 0xff);
                size >>= 8;
            }
        }
        Bp[8] = (cbyte)(h[i].flags >> 8);
        Bp[9] = (cbyte)(h[i].flags & 0xff);
        memcpy(&Bp[10], h[i].value, h[i].size);

        Bp += h[i].size + 10; /* +10 for size of frame header */
    }
    assert(Bp == B+id3_size+10);

    // That's the new header in place.  Now, what do we do with it?

    if (new_filename == NULL) {
        // there's space -- replace the contents of the existing header
        if (verbose) {
            fprintf(stderr, "Replacing %zd bytes in file\n", id3_size);
        }
        char* oldB = (char*)id3->base;
        size_t start_of_padding = 10+id3_size;
        size_t size_of_padding  = header_size - id3_size - 10;
        
        memcpy(oldB+10, B+10, id3_size);
        memset(oldB+start_of_padding, 0, size_of_padding);

        // no other bookkeeping is required -- the contents of *id3
        // are still good
        return_mp3 = id3;
    } else {
        // create a new file
        int fd;
        size_t new_filesize;
        size_t new_buflen;
        char* newB;
        char* Bcursor;
        

        new_filesize = 10 + id3_size + new_padding_size + body_size;
        if (verbose) {
            fprintf(stderr, "Creating new file %s of size %zd bytes\n", new_filename, new_filesize);
        }

        if ((fd = open(new_filename, O_CREAT|O_RDWR, 0666)) < 0) {
            fprintf(stderr, "%s: can't open file %s to write [%s]\n",
                    program_name(), new_filename, strerror(errno));
            exit(1);
        }
        if (ftruncate(fd, new_filesize) < 0) {
            fprintf(stderr, "%s: can't truncate file %s to length %zd [%s]\n",
                    program_name(), new_filename, new_filesize, strerror(errno));
            exit(1);
        }

        if ((newB = (char*)map_fd(fd, &new_buflen, 1)) == NULL) {
            fprintf(stderr, "%s: unable to map newly-truncated file [%s]\n",
                    program_name(), strerror(errno));
            exit(1);
        }
        assert(new_buflen == new_filesize);

        Bcursor = newB;
        
        // Copy ID3 header
        memcpy(Bcursor, B, id3_size + 10);
        Bcursor += id3_size + 10;

        // Add padding bytes
        memset(Bcursor, 0, new_padding_size);
        Bcursor += new_padding_size;

        // Add audio
        memcpy(Bcursor, id3->audio, body_size);

        release_mp3((MP3Info*)id3);

        return_mp3 = initialise_mp3((cbyte*)newB, new_buflen, new_filename);
    }    

    *actual_new_filename = new_filename;
    
    return return_mp3;
}

static struct id3v2_header* add_header(MP3Info* i)
{
    struct id3v2_header* h;

    if (i->nmetadata == i->metadata_size) {
        // NB: this _requires_ that i is initialised with
        // nmetadata == metadata_size == 0
        if (i->metadata_size == 0) {
            // first time
            i->metadata_size = 8; /* fairly arbitrary */
        } else {
            i->metadata_size *= 2;
        }
        i->metadata = (struct id3v2_header*)realloc(i->metadata,
                                                    i->metadata_size * sizeof(struct id3v2_header));
        if (i->metadata == NULL) {
            fprintf(stderr, "read_id3_headers: Error allocating memory\n");
            exit(1);
        }
    }

    h = &(i->metadata[i->nmetadata]);
    (i->nmetadata)++;
    return h;
}
