/*
 * This is two evenings' attempt at a bare-bones MP3 frame scanner
 * (well, it's more than two evenings now, but I got the bones of it
 * working quite quickly).
 * The goal was to produce something that would report the length of
 * an MP3 in milliseconds
 *
 * MPEG-1 is described in ISO/IEC 11172-3.
 * MPEG-2 is described in ISO/IEC 13818-3.
 * An overview of the structure, including a mention of MPEG-2 v2.5,
 * are in http://www.mpgedit.org/mpgedit/mpeg_format/mpeghdr.htm
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <assert.h>
#include <math.h>

#include <config.h>

#include "id3.h"
#include "util.h"

const char *progname;


typedef struct {
    unsigned int bitrate;
    unsigned int sample_rate;
    unsigned int version;
    unsigned int layer;
    unsigned int mode;
    unsigned int mode_extn;
    int length;
    float duration_ms;
} MP3_frame_info;

static cbyte* buffer_start_for_debug;
#define OFS(p) (size_t)((p)-buffer_start_for_debug)

/* MPEG definitions */
#define MPEG25 0
#define MPEG_BADVERSION 1
#define MPEG2  2
#define MPEG1  3

#define MPEG_BADLAYER 0
#define MPEG_LAYER3 1
#define MPEG_LAYER2 2
#define MPEG_LAYER1 3

#define MPEG_STEREO 0
#define MPEG_JOINTSTEREO 1
#define MPEG_DUALSTEREO 2
#define MPEG_MONO 3

/* Printable versions */
const char* mp3_version[] = { "v2.5",             // 00
                              "BAD",              // 01
                              "v2",               // 10
                              "v1" };             // 11
const char* mp3_layer[]   = { "BAD",              // 00
                              "layer 3",          // 01
                              "layer 2",          // 10
                              "layer 1" };        // 11
const char* mp3_mode[]    = { "Stereo",           // 00
                              "Joint stereo",     // 01
                              "Dual channel",     // 10
                              "Single channel" }; // 11
// joint stereo modes: ISO-11172-3 Sect 2.4.2.3
const char* mp3_mode_extn_12[] = { "4",
                                   "8",
                                   "12",
                                   "16" };
const char* mp3_mode_extn_3[] = { "off/off",
                                  "inten/off",
                                  "off/ms",
                                  "inten/ms" };    

/* Bitrate lookup mp3_bitrate[version][layer] */
const int mp3_bitrate[8][16] = {
    { 0,  0,  0,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 0 }, // dummy
    { 0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 0 }, // v1, layer 3
    { 0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, 0 }, // v1, layer 2
    { 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0 }, // v1, layer 1
    { 0,  0,  0,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 0 }, // dummy
    { 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 }, // v2, layer 3
    { 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 }, // v2, layer 2
    { 0, 32, 48, 56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256, 0 }, // v2, layer 1
};

// mp3_sampling_rates[version][frequency-index]
const int mp3_sampling_rate[4][4] = {
    { 11025, 12000,  8000, 0 }, // MPEG v2.5
    {     0,     0,     0, 0 }, // invalid
    { 22050, 24000, 16000, 0 }, // MPEG-2
    { 44100, 48000, 32000, 0 }, // MPEG-1
};

/* function definitions */
static void Usage(void);
static const MP3Info* open_mp3file(const char* filename,
                                   int read_write_p);
static const MP3Info* examine_buffer(const MP3Info *info,
                                     unsigned int flags);
static cbyte* scan_for_frame(cbyte* p, cbyte* max);
static MP3_frame_info* parse_frame(cbyte* p, int scanning_for_sync);
static const MP3Info* display_id3v2(const MP3Info *info,
                                    unsigned int display_flags);

typedef struct new_tag {
    const char* tag;
    const char* value;
    struct new_tag* next;
} OneTag;


#define FLAG_SHOW_FRAMES 1
#define FLAG_SHOW_DURATION 2
#define FLAG_SHOW_DURATION_HMS 4
#define FLAG_SHOW_FILENAMES 8
#define FLAG_SHOW_ID3 16
#define FLAG_EDIT_ID3 32
#define FLAG_SHOW_VERSION 64
#define FLAG_DEFAULT (FLAG_SHOW_FRAMES | FLAG_SHOW_DURATION | FLAG_SHOW_FILENAMES | FLAG_SHOW_ID3)

#ifndef HAVE_LROUNDF
// I test for this, but I haven't yet tried to build on a platform
// that doesn't have this, so haven't had to implement it yet.
#error "Can't find lroundf -- implement it"
#endif
#if !HAVE_DECL_LROUNDF
// Oddly, SL4.8 (for example) has lroundf but doesn't declare it
long lroundf(float x);
#endif

int verbose = 0;
#define VERBOSE_QUIET 0
#define VERBOSE_CHATTY 1
#define VERBOSE_DETAILS 2
#define VERBOSE_DEBUG 3

int main(int argc, char **argv) 
{
    int ch;
    unsigned int flags = 0;
    OneTag *tags = NULL;
    char *output_filename = NULL;
    
    progname = argv[0];
    while ((ch = getopt(argc, argv, "e:fhino:tTvV")) != -1) {
        switch (ch) {
          case 'e':
            {
                OneTag *n;
                const char* value = optarg;
                // Require that optarg is of the form Txxx=value
                if (value == NULL
                    || strlen(value) < 5
                    || value[4] != '=') {
                    fprintf(stderr, "Unexpected argument for -e: <%s>\n", value);
                    Usage();
                }
                if ((n = (OneTag*)malloc(sizeof(OneTag))) == NULL) {
                    fprintf(stderr, "Can't allocate new tag struct!\n");
                    exit(1);
                }
                n->next = NULL;
                n->tag = value;
                if (value[5] == '\0') {
                    n->value = NULL; // indicate that a tag is to be removed
                } else {
                    n->value = &value[5];
                }
                // add to the end of the list pointed to by 'tags'
                if (tags == NULL) {
                    tags = n;
                } else {
                    OneTag *t = tags;
                    while (t->next != NULL) {
                        t = t->next;
                    }
                    assert(t->next == NULL);
                    t->next = n;
                }
                flags |= FLAG_EDIT_ID3;
                break;
            }
          case 'f':
            flags |= FLAG_SHOW_FILENAMES;
            break;
          case 'h':
            Usage();
            break;
          case 'i':
            flags |= FLAG_SHOW_ID3;
            break;
          case 'n':
            flags |= FLAG_SHOW_FRAMES;
            break;
          case 'o':
            output_filename = optarg;
            break;
          case 't':
            flags |= FLAG_SHOW_DURATION;
            break;
          case 'T':
            flags |= FLAG_SHOW_DURATION_HMS;
            break;
          case 'v':
            verbose++;
            break;
          case 'V':
            flags |= FLAG_SHOW_VERSION;
            break;
          case '?':
            Usage();
          default:
            fprintf(stderr, "Impossible option! %c\n", ch);
            exit(1);
        }
    }
    argc -= optind;
    argv += optind;
    if (argc < 1) {
        if (! (flags & FLAG_SHOW_VERSION)) {
            Usage();
        }
    }

    if (flags == 0) {
        flags = FLAG_DEFAULT;
    }

    if (flags & FLAG_SHOW_VERSION) {
        printf("mp3x version %s, released %s; see also <%s>.\n",
               PACKAGE_VERSION, RELEASEDATE, PACKAGE_URL);
    }

    for (; argc > 0; argc--, argv++) {
        const char *filename = *argv;
        const MP3Info* id3_info;

        id3_info = open_mp3file(filename, (flags & FLAG_EDIT_ID3));
        if (id3_info == NULL) {
            fprintf(stderr, "%s: can't open MP3 file %s\n", progname, filename);
            exit(1);
        }

        if (verbose >= VERBOSE_CHATTY) {
            printf("File %s; %zd bytes of audio", id3_info->filename, (id3_info->audio_end - id3_info->audio));
            if (id3_info->base != id3_info->audio) {
                printf(" in file of %zd bytes", (id3_info->audio_end - id3_info->base));
            }
            if (id3_info->metadata == NULL) {
                printf("\n(no file metadata)\n");
            } else {
                printf("\nID3 header: %zd bytes, of which %zd is padding (ID3v2.%d.%d, flags=0x%x)\n",
                       (id3_info->audio - id3_info->base), id3_info->padding,
                       id3_info->major_version, id3_info->minor_version, id3_info->flags);
            }
        }

        id3_info = examine_buffer(display_id3v2(id3_info, flags), flags);

        if (flags & FLAG_EDIT_ID3) {
            OneTag* p;
            const char *actual_output_filename;
            
            assert(tags != NULL);

            for (p=tags; p!=NULL; p=p->next) {
                add_id3_header_text((MP3Info*)id3_info, /* lose const */
                                    p->tag,
                                    p->value);
            }
            id3_info = write_header_and_file(id3_info,
                                             output_filename,
                                             &actual_output_filename);

            if (output_filename != NULL) {
                if (actual_output_filename != NULL) {
                    if (strcmp(actual_output_filename, output_filename) != 0) {
                        // output to different filename
                        // FIXME: this won't work if it crosses filesystems
                        rename(actual_output_filename, output_filename);
                    } else {
                        // output filename is what we wanted -- nothing to do
                    }
                } else {
                    // something went wrong
                    fprintf(stderr, "%s: error writing resulting MP3 file\n", progname);
                }
            } else {
                // output_filename == NULL
                // we wanted to replace the file
                if (actual_output_filename == NULL) {
                    // and we did so -- nothing to do
                } else {
                    // FIXME: this won't work if it crosses filesystems
                    rename(actual_output_filename, filename);
                }
            }
        }

        release_mp3((MP3Info*)id3_info);
    }

    exit(0);
}

static MP3_frame_info* parse_frame(cbyte* p, int scanning_for_sync)
{
    static MP3_frame_info info;
    MP3_frame_info* return_value = 0;

    if (p[0] == 0xff && (p[1] & 0xe0) == 0xe0) {
        // first 11 bits set

        // aaaa aaaa | aaab bccd | eeee ffgh | iijj klmm
        // a = sync
        // b = version
        // c = layer
        // d = protection
        // e = bitrate
        // f = sampling
        // g = padding
        // h = private
        // i = channel mode
        // j = mode extension
        // k = copyright
        // l = original
        // m = emphasis
        int h_version    = (p[1] & 0x18) >> 3;
        int h_layer      = (p[1] & 0x06) >> 1;
        int h_protection =  p[1] & 0x01;
        int h_bitrate    = (p[2] & 0xf0) >> 4;
        int h_sampling   = (p[2] & 0x0c) >> 2;
        int h_padding    = (p[2] & 0x02) >> 1;
        int h_mode       = (p[3] & 0xc0) >> 6;
        int h_mode_extn  = (p[3] & 0x30) >> 4;

        info.sample_rate = mp3_sampling_rate[h_version][h_sampling];
        info.bitrate = mp3_bitrate[(h_version==MPEG1 ? 0 : 4) + h_layer][h_bitrate];
        info.version = h_version;
        info.layer = h_layer;
        info.mode = h_mode;
        info.mode_extn = h_mode_extn;

        if (verbose >= VERBOSE_DEBUG) {
#if 1
            printf("?frame at %lx: %02x %02x %02x %02x: %s/%s, %s %d/%d; crc? %s, padding? %s\n",
                   OFS(p), p[0], p[1], p[2], p[3],
                   mp3_version[h_version],
                   mp3_layer[h_layer],
                   mp3_mode[h_mode],
                   info.bitrate,
                   info.sample_rate,
                   (h_protection ? "no" : "yes"),
                   (h_padding ? "yes" : "no"));
#else
            printf("    version=%d(%s)  layer=%d(%s)  mode=%d(%s)  bitrate=%x(%d)  sampling=%x(%d)  has crc? %s  padding? (%d)%s\n",
                   h_version, mp3_version[h_version],
                   h_layer,  mp3_layer[h_layer],
                   h_mode, mp3_mode[h_mode],
                   h_bitrate, info.bitrate,
                   h_sampling, info.sample_rate,
                   (h_protection ? "no" : "yes"),
                   h_padding, (h_padding ? "yes" : "no"));
#endif
        }

        /*
         * When we're scanning_for_sync, these are tests to
         * corroborate that we have in fact found a valid frame
         * header.  When we're parsing the file later, these are
         * only sanity checks.
         */
        if (h_version == MPEG_BADVERSION) {
            if (verbose >= VERBOSE_DEBUG)
                printf("Bad version %d\n", h_version);
        } else if (h_layer == MPEG_BADLAYER) {
            if (verbose >= VERBOSE_DEBUG)
                printf("Bad layer %d\n", h_layer);
        } else if (info.bitrate == 0 || info.sample_rate == 0) {
            if (verbose >= VERBOSE_DEBUG)
                printf("Bad bitrate or sample rate\n");
        } else {
            /*
             * The MPEG-1 length calculation is described in
             * 11172, Sects. 2.4.2 & 2.4.3.1.  It's the same for
             * MPEG-2, apparently, but....
             *
             * 11172 Sect 2.4.3.1 indicates the calculation for
             * the number of slots per frame, in the two levels,
             * and this implies the 48(=12*4) or 144 below.  The
             * number 144 is 1152 samples divided by 8 (I don't
             * know where the 8 comes from).
             *
             * 13818 seems to confirm this, giving the number 144
             * in C.1.2.  However Sect 2.5.2.1 of that version
             * states that there are _either_ 1152 or 578
             * samples per frame 'for each multilingual channel',
             * and 2.5.2.14 and .15 talk about halving the
             * sampling frequency if 'multi_lingual_fs==1'.  But
             * the only place that multi_lingual_fs is specified
             * is within mc_header, and the only way to find the
             * start of that is to decode the audio_data, and the
             * only way to do that is to go through it byte by
             * byte (it appears).
             *
             * The notes at http://www.codeproject.com/KB/audio-video/mpegaudioinfo.aspx
             * simply state that there are 576 samples per frame
             * in MPEG-2/III (and v2.5), without qualification and
             * with no explanation.  This appears to work for the
             * small number of 2/III files I've tried it with.
             *
             * Perhaps a way to work this out (in the case of
             * MPEG-2/III or MPEG-2.5/III) is to examine the file
             * assuming a multiplier of 72, and if there's a sync
             * header there, take that; if not, examine it using a
             * multiplier of 144.  That would require setting some
             * global status.
             */
            if (h_layer == MPEG_LAYER1) {
                // four-byte slots
                info.length = (12 * info.bitrate * 1000 / info.sample_rate + h_padding) * 4;
                // duration in ms is 1/sample-rate * 384 samples * 1000ms
                info.duration_ms = 384000.0 / info.sample_rate;
            } else {
                // one-byte slots (144 = 1152samples/8)
                int sample_multiplier;
                float time_multiplier;
                if ((h_version == MPEG2 || h_version == MPEG25) && h_layer == MPEG_LAYER3) {
                    // MPEG-2/III or MPEG-2.5/III
                    sample_multiplier = 72;
                    time_multiplier = 576000.0;
                } else {
                    sample_multiplier = 144;
                    // duration in ms is 1/sample-rate * 1152 samples * 1000ms
                    time_multiplier = 1152000.0;
                }
                info.length = sample_multiplier * info.bitrate * 1000 / info.sample_rate + h_padding;
                info.duration_ms = time_multiplier / info.sample_rate;
            }
            if (scanning_for_sync) {
                // when scanning at the start, the final test is to check that the 
                // byte where the next frame should be does indeed look like a sync byte.
                if (verbose >= VERBOSE_DEBUG)
                    printf("    length = %x: bytes at %lx = %x %x %x %x\n",
                           info.length, OFS(p+info.length),
                           p[info.length], p[info.length+1], p[info.length+2], p[info.length+3]);
                if (p[info.length] == 0xff) {
                    // looks like a frame
                    return_value = &info;
                }
            } else {
                // passed all the tests so far, so it's a valid frame, return it
                return_value = &info;
            }
        }
    }

    // output debugging information
    if (return_value == NULL && verbose >= VERBOSE_DEBUG) {
        printf("Unexpected frame @ %lx...? frame= %x %x %x %x\n", OFS(p), p[0], p[1], p[2], p[3]);
    }

    return return_value;
}

static cbyte* scan_for_frame(cbyte* start, cbyte* buffer_end)
{
    int nscanned;
    MP3_frame_info* info = 0;
    cbyte* looks_like_frame = 0;
    int n_to_scan = 4;          /* number of frames to check for consistency */
    cbyte* p = start;

    for (nscanned=0; p < buffer_end; p++, nscanned++) {
        if ((info = parse_frame(p, 1)) == NULL) {
            looks_like_frame = 0; /* this isn't a frame (keep scanning) */
        } else {
            int scanned;
            looks_like_frame = p; /* candidate frame -- look ahead to check */
            for (scanned = 0; scanned<n_to_scan; scanned++) {
                if (verbose >= VERBOSE_DEBUG)
                    printf("Lookahead %d...\n", scanned);
                p += info->length; /* this should be a frame, too */
                if ((info = parse_frame(p, 1)) == NULL) {
                    // this isn't a frame, after all
                    p = looks_like_frame;
                    break;
                }
            }
            if (scanned == n_to_scan) {
                // we worked through n_to_scan frame, and they were all valid, so...
                if (verbose >= VERBOSE_DEBUG)
                    printf("I think a frame starts back at %lx\n", OFS(looks_like_frame));
                break;
            }
        }
    }
    return looks_like_frame;
}

static const MP3Info* open_mp3file(const char* filename,
                                   int read_write_p)
{
    cbyte* buf;
    size_t buflen;
                
    if ((buf = map_file(filename, &buflen, read_write_p)) == NULL) {
        fprintf(stderr, "%s: can't map file %s\n", progname, filename);
        exit(1);
    }

    if (verbose >= VERBOSE_DEBUG) {
        int i;
        printf("File %s", filename);
        for (i=0; i<10; i++)
            printf(" %x", buf[i]);
        putchar('\n');
    }

    read_id3_header_verbosity(verbose);
    return initialise_mp3(buf, buflen, filename);
}

static const MP3Info* examine_buffer(const MP3Info *id3_info,
                                     unsigned int display_flags)
{
    cbyte* p;
    cbyte* buffer_end = id3_info->audio_end;
    MP3_frame_info* frame_info;
    int frame_count;
    float duration_ms;

    if (id3_info == NULL) {
        return NULL;
    }
    
    buffer_start_for_debug = id3_info->base; /* initialise debugging function, so OFS(p) works */

    p = scan_for_frame(id3_info->audio, buffer_end);
    if (p == NULL) {
        fprintf(stderr, "No sync frames found!\n");
        exit(1);                /* EXIT */
    } else {
        if (verbose >= VERBOSE_CHATTY)
            printf("scan_for_frame: found at 0x%lx\n", OFS(p));

        // start scanning the frames
        frame_count = 0;
        duration_ms = 0;

        // This is the normal main loop of the function
        // (and indeed of the program)
        for (; p<buffer_end; p+=frame_info->length) {
            
            frame_info = parse_frame(p, 0);
            if (frame_info == NULL) {
                // we're going to exit the loop
                int error_exit = 0; // ...or exit the program
                
                if ((buffer_end - p == 128) && strncmp((const char*)p, "TAG", 3) == 0) {
                    // This is a (trailing) ID3 tag (cf http://www.id3.org/ID3v1)
                    if (verbose >= VERBOSE_CHATTY) {
                        printf("Found ID3v1 tag\n");
                    }
                } else {
                    // that's not supposed to happen
                    if (buffer_end - p < 0x1000) {
                        // ...but we're pretty close to the end, so
                        // this is presumably just a botched ID3v1
                        // tag, which we're not going to worry about
                        p = buffer_end;
                    } else {
                        // something's gone wrong -- give up
                        fprintf(stderr, "Unexpected loss of sync at %ld (%lx)\n", OFS(p), OFS(p));
                        error_exit = 1;
                    }
                }

                if (error_exit) {
                    exit(1);    /* EXIT */
                }

                break;      /* JUMP OUT of loop */
            }

            frame_count++;
            duration_ms += frame_info->duration_ms;
            if (verbose >= VERBOSE_DETAILS) {
                // Show information.  Final duration is the sum of the
                // durations to the end of this frame.
                printf("Frame %4d @ %4lx: %s/%s, %s %s %3d/%d, +%gms\n",
                       frame_count, OFS(p),
                       mp3_version[frame_info->version],
                       mp3_layer[frame_info->layer],
                       mp3_mode[frame_info->mode],
                       (frame_info->mode == 1
                        ? (frame_info->layer == MPEG_LAYER3
                           ? mp3_mode_extn_3[frame_info->mode_extn]
                           : mp3_mode_extn_12[frame_info->mode_extn])
                        : ""),
                       frame_info->bitrate,
                       frame_info->sample_rate, frame_info->duration_ms);
                if (verbose >= VERBOSE_DEBUG) {
                    printf("    so step=%x bytes, cumulative=%gms\n",
                           frame_info->length, duration_ms);
                }
            }
            assert(frame_info->length > 0);
        }
        
        if (verbose >= VERBOSE_CHATTY) {
            printf("Total: %d frames, %ldms\n", frame_count, lroundf(duration_ms));
        }

        if (display_flags == 0)
            display_flags = FLAG_DEFAULT;

        {
            int output_tab = 0;
#define TAB { if (output_tab) putchar('\t'); output_tab = 1; }

            if (display_flags & FLAG_SHOW_FRAMES) {
                TAB;
                printf("%d", frame_count);
            }
            if (display_flags & FLAG_SHOW_DURATION) {
                TAB;
                printf("%ld", lroundf(duration_ms));
            }
            if (display_flags & FLAG_SHOW_DURATION_HMS) {
                int tot_ms = lroundf(duration_ms);
                int ms = tot_ms % 1000;
                int s = tot_ms / 1000;
                int m = s / 60;
                int h = m / 60;
                s = s - m*60;
                m = m - h*60;
                TAB;
                if (h == 0) {
                    printf("%d:%02d.%d", m, s, ms);
                } else {
                    printf("%d:%02d:%02d.%d", h, m, s, ms);
                }
            }
            if (display_flags & FLAG_SHOW_FILENAMES) {
                TAB;
                printf("%s", id3_info->filename);
            }
            if (output_tab) {
                putchar('\n');
            }
#undef TAB
        }
    }

    return id3_info;
}

// Display the contents of the ID3v2 header, starting at p.
// Return a pointer one byte past the end of the buffer.
// If the display_flags do not require that we parse the ID3 headers,
// then simply skip over the ID3 stuff and set the end-of-ID3 pointer.
static const MP3Info* display_id3v2(const MP3Info *info,
                                    unsigned int display_flags)
{
    int read_tags_p;
    
    if (info == NULL) {
        return NULL;
    }
    assert(info != NULL);

    read_tags_p = (display_flags & (FLAG_SHOW_ID3 | FLAG_EDIT_ID3));

    if (read_tags_p) {
        if (display_flags & FLAG_SHOW_ID3) {
            int hn, i;
            struct id3v2_header* h;

            h = info->metadata;
            hn = info->nmetadata;
            for (i=0; i<hn; i++) {
                print_id3_frame(&h[i], stdout);
            }
        }
    }

    return info;
}

// Make the program name available to other modules
const char* program_name() 
{
    return progname;
}

static void Usage(void)
{
    fprintf(stderr, "Usage: %s [-fhintTv] [-e Txxx=value] [-o file] filename ...\n\
\n\
Display options:\n\
    -f : show filename\n\
    -h : display this help\n\
    -i : show ID3 info\n\
    -n : show number of frames\n\
    -t : show MP3 duration in ms\n\
    -T : show MP3 duration in h:m:s\n\
    -v : verbose output (-vv more verbose)\n\
Processing options\n\
    -e : add or change ID3v2 tag, or delete one if 'value' is absent\n\
    -o : set output filename\n\
    -V : report version information\n\
\n\
The ID3 information is shown as s sequence of 'TAGS=value' lines.\n\
The number of frames, duration and filename are shown on a single line,\n\
if present, in that order.\n\
\n\
If no flags are given on the command line, acts as if -intf were present.\n", progname);
    exit(1);
}
